/* :nodoc: */
#include "bf_measurement.h"

#if defined(__APPLE__)
#include <mach/mach_time.h>
#elif !defined(_WIN32)
#include <time.h>
#endif

static VALUE cMeasure;
VALUE total_allocated_objects_key;
;

static struct four_doubles_t measurers(void *measurer_ptr, rb_trace_arg_t *trace_arg)
{
    struct four_doubles_t ret = {0, 0, 0, 0};
    prof_measurer_t *measurer = (prof_measurer_t *)measurer_ptr;

    /* wall time */
#if defined(_WIN32)
    LARGE_INTEGER time;
    QueryPerformanceCounter(&time);
    ret.wt = time.QuadPart;
#elif defined(__APPLE__)
    ret.wt = mach_absolute_time();
#elif defined(__linux__)
    struct timespec tv;
    clock_gettime(CLOCK_MONOTONIC, &tv);
    ret.wt = tv.tv_sec + (tv.tv_nsec / 1000000000.0);
#else
    struct timeval tv;
    gettimeofday(&tv, NULL);
    ret.wt = tv.tv_sec + (tv.tv_usec / 1000000.0);
#endif

    if (measurer->flag_cpu) {
    /* process time */
#if defined(_WIN32)
        FILETIME createTime;
        FILETIME exitTime;
        FILETIME sysTime;
        FILETIME userTime;
        ULARGE_INTEGER sysTimeInt;
        ULARGE_INTEGER userTimeInt;

        GetProcessTimes(GetCurrentProcess(), &createTime, &exitTime, &sysTime, &userTime);

        sysTimeInt.LowPart = sysTime.dwLowDateTime;
        sysTimeInt.HighPart = sysTime.dwHighDateTime;
        userTimeInt.LowPart = userTime.dwLowDateTime;
        userTimeInt.HighPart = userTime.dwHighDateTime;

        ret.pr = sysTimeInt.QuadPart + userTimeInt.QuadPart;
#elif !defined(CLOCK_PROCESS_CPUTIME_ID)
        struct rusage usage;
        getrusage(RUSAGE_SELF, &usage);
        ret.pr = usage.ru_stime.tv_sec + usage.ru_utime.tv_sec + ((usage.ru_stime.tv_usec + usage.ru_utime.tv_usec) / 1000000.0);
#else
        struct timespec clock;
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &clock);
        ret.pr = clock.tv_sec + (clock.tv_nsec / 1000000000.0);
#endif
    }

    /* memory */
    static double result = 0;

    if (trace_arg)
    {
        rb_event_flag_t event = rb_tracearg_event_flag(trace_arg);
        if (event == RUBY_INTERNAL_EVENT_NEWOBJ)
        {
            VALUE object = rb_tracearg_object(trace_arg);
            result += rb_obj_memsize_of(object);
        }
    }
    ret.mm = result;

    /* allocations */
    if (total_allocated_objects_key)
    {
        ret.al = rb_gc_stat(total_allocated_objects_key);
    }
    else
    {
        static double resultal = 0;

        if (trace_arg)
        {
            rb_event_flag_t event = rb_tracearg_event_flag(trace_arg);
            if (event == RUBY_INTERNAL_EVENT_NEWOBJ)
            {
                resultal++;
            }
        }

        ret.al = resultal;
    }

    return ret;
}

/*
 *
 */
static struct four_doubles_t multipliers(void)
{
    struct four_doubles_t ret;

    /* wall time */
#if defined(_WIN32)
    LARGE_INTEGER frequency;
    QueryPerformanceFrequency(&frequency);
    ret.wt = 1.0 / frequency.QuadPart;
#elif defined(__APPLE__)
    mach_timebase_info_data_t mach_timebase;
    mach_timebase_info(&mach_timebase);
    ret.wt = (uint64_t)mach_timebase.numer / (uint64_t)mach_timebase.denom / 1000000000.0;
#else
    ret.wt = 1.0;
#endif

    /* process time */
#if defined(_WIN32)
    // Times are in 100-nanosecond time units.  So instead of 10-9 use 10-7
    ret.pr = 1.0 / 10000000.0;
#else
    ret.pr = 1.0;
#endif

    ret.mm = 1.0;
    ret.al = 1.0;

    return ret;
}

prof_measurer_t *prof_measurer_wall_process(void)
{
    prof_measurer_t *measure = ALLOC(prof_measurer_t);
    measure->measure = measurers;
    measure->multipliers = multipliers();
    return measure;
}

void bf_init_measure() { cMeasure = rb_define_class_under(mMeasure, "Measure", rb_cObject); }
