/* Copyright (C) 2005-2019 Shugo Maeda <shugo@ruby-lang.org> and Charlie Savage
   <cfis@savagexi.com> Please see the LICENSE file for copyright and
   distribution information */

#include "bf_stack.h"

#define INITIAL_STACK_SIZE 64

// Creates a stack of prof_frame_t to keep track of timings for active methods.
prof_stack_t *prof_stack_create()
{
    prof_stack_t *stack = ALLOC(prof_stack_t);
    stack->start = ZALLOC_N(prof_frame_t, INITIAL_STACK_SIZE);
    stack->ptr = stack->start;
    stack->end = stack->start + INITIAL_STACK_SIZE;

    return stack;
}

void prof_stack_free(prof_stack_t *stack)
{
    xfree(stack->start);
    xfree(stack);
}

prof_frame_t *prof_stack_last(prof_stack_t *stack)
{
    if (stack->ptr == stack->start)
    {
        return NULL;
    }
    return stack->ptr - 1;
}

void prof_stack_verify_size(prof_stack_t *stack)
{
    // Is there space on the stack?  If not, double its size.
    if (stack->ptr == stack->end)
    {
        size_t len = stack->ptr - stack->start;
        size_t new_capacity = (stack->end - stack->start) * 2;
        REALLOC_N(stack->start, prof_frame_t, new_capacity);

        /* Memory just got moved, reset pointers */
        stack->ptr = stack->start + len;
        stack->end = stack->start + new_capacity;
    }
}

prof_frame_t *prof_stack_push(prof_stack_t *stack)
{
    prof_stack_verify_size(stack);

    prof_frame_t *result = stack->ptr;
    stack->ptr++;
    return result;
}

prof_frame_t *prof_stack_pop(prof_stack_t *stack)
{
    prof_frame_t *result = prof_stack_last(stack);
    if (result)
    {
        stack->ptr--;
    }

    return result;
}

// ----------------  Frame Methods  ----------------------------
void prof_frame_pause(prof_frame_t *frame, struct four_doubles_t current_measurement)
{
    if (frame && prof_frame_is_unpaused(frame))
    {
        frame->pause_time = current_measurement;
    }
}

void prof_frame_unpause(prof_frame_t *frame, struct four_doubles_t current_measurement)
{
    if (prof_frame_is_paused(frame))
    {
        frame->dead_time.wt += (current_measurement.wt - frame->pause_time.wt);
        frame->dead_time.mm += (current_measurement.mm - frame->pause_time.mm);
        frame->dead_time.al += (current_measurement.al - frame->pause_time.al);
        frame->dead_time.pr += (current_measurement.pr - frame->pause_time.pr);
        frame->pause_time.wt = -1;
        frame->pause_time.mm = -1;
        frame->pause_time.pr = -1;
        frame->pause_time.al = -1;
    }
}

prof_frame_t *prof_frame_current(prof_stack_t *stack) { return prof_stack_last(stack); }

prof_frame_t *prof_frame_push(prof_stack_t *stack, prof_call_tree_t *call_tree, struct four_doubles_t measurement, bool paused)
{
    prof_frame_t *parent_frame = prof_stack_last(stack);
    prof_frame_t *result = prof_stack_push(stack);

    result->call_tree = call_tree;

    result->start_time = measurement;

    result->pause_time.wt = -1;
    result->pause_time.mm = -1;
    result->pause_time.pr = -1;
    result->pause_time.al = -1;

    result->switch_time.wt = 0;
    result->switch_time.mm = 0;
    result->switch_time.pr = 0;
    result->switch_time.al = 0;

    result->wait_time.wt = 0;
    result->wait_time.mm = 0;
    result->wait_time.pr = 0;
    result->wait_time.al = 0;

    result->child_time.wt = 0;
    result->child_time.mm = 0;
    result->child_time.pr = 0;
    result->child_time.al = 0;

    result->dead_time.wt = 0;
    result->dead_time.mm = 0;
    result->dead_time.pr = 0;
    result->dead_time.al = 0;

    result->source_file = Qnil;
    result->source_line = 0;

    call_tree->measurement->called++;
    call_tree->visits++;

    if (call_tree->method->visits > 0)
    {
        call_tree->method->recursive = true;
    }
    call_tree->method->measurement->called++;
    call_tree->method->visits++;

    // Unpause the parent frame, if it exists.
    // If currently paused then:
    //   1) The child frame will begin paused.
    //   2) The parent will inherit the child's dead time.
    if (parent_frame)
    {
        prof_frame_unpause(parent_frame, measurement);
    }

    if (paused)
    {
        prof_frame_pause(result, measurement);
    }

    // Return the result
    return result;
}

prof_frame_t *prof_frame_unshift(prof_stack_t *stack, prof_call_tree_t *parent_call_tree, prof_call_tree_t *call_tree, struct four_doubles_t measurement)
{
    if (prof_stack_last(stack))
    {
        rb_raise(rb_eRuntimeError, "Stach unshift can only be called with an empty stack");
    }

    parent_call_tree->measurement->wt_total_time = call_tree->measurement->wt_total_time;
    parent_call_tree->measurement->wt_self_time = 0;
    parent_call_tree->measurement->wt_wait_time = call_tree->measurement->wt_wait_time;

    parent_call_tree->measurement->pr_total_time = call_tree->measurement->pr_total_time;
    parent_call_tree->measurement->pr_self_time = 0;
    parent_call_tree->measurement->pr_wait_time = call_tree->measurement->pr_wait_time;

    parent_call_tree->measurement->al_total_time = call_tree->measurement->al_total_time;
    parent_call_tree->measurement->al_self_time = 0;
    parent_call_tree->measurement->al_wait_time = call_tree->measurement->al_wait_time;

    parent_call_tree->measurement->mm_self_time = 0;
    parent_call_tree->measurement->mm_wait_time = call_tree->measurement->mm_wait_time;

    parent_call_tree->method->measurement->wt_total_time += call_tree->measurement->wt_total_time;
    parent_call_tree->method->measurement->wt_wait_time += call_tree->measurement->wt_wait_time;

    parent_call_tree->method->measurement->pr_total_time += call_tree->measurement->pr_total_time;
    parent_call_tree->method->measurement->pr_wait_time += call_tree->measurement->pr_wait_time;

    parent_call_tree->method->measurement->mm_wait_time += call_tree->measurement->mm_wait_time;

    parent_call_tree->method->measurement->al_total_time += call_tree->measurement->al_total_time;
    parent_call_tree->method->measurement->al_wait_time += call_tree->measurement->al_wait_time;

    return prof_frame_push(stack, parent_call_tree, measurement, false);
}

prof_frame_t *prof_frame_pop(prof_stack_t *stack, struct four_doubles_t measurement, prof_measurer_t *measurer)
{
    prof_frame_t *frame = prof_stack_pop(stack);

    if (!frame)
    {
        return NULL;
    }

    /* Calculate the total time this method took */
    prof_frame_unpause(frame, measurement);

    double total_time_wt = measurement.wt - frame->start_time.wt - frame->dead_time.wt;
    double self_time_wt = total_time_wt - frame->child_time.wt - frame->wait_time.wt;
    double total_time_pr = measurement.pr - frame->start_time.pr - frame->dead_time.pr;
    double self_time_pr = total_time_pr - frame->child_time.pr - frame->wait_time.pr;
    double total_time_mm = measurement.mm - frame->start_time.mm - frame->dead_time.mm;
    double self_time_mm = total_time_mm - frame->child_time.mm - frame->wait_time.mm;
    double total_time_al = measurement.al - frame->start_time.al - frame->dead_time.al;
    double self_time_al = total_time_al - frame->child_time.al - frame->wait_time.al;

    /* Update information about the current method */
    prof_call_tree_t *call_tree = frame->call_tree;

    call_tree->method->measurement->wt_self_time += self_time_wt;
    call_tree->method->measurement->wt_wait_time += frame->wait_time.wt;
    call_tree->measurement->wt_self_time += self_time_wt;
    call_tree->measurement->wt_wait_time += frame->wait_time.wt;
    if (call_tree->method->visits == 1)
    {
        call_tree->method->measurement->wt_total_time += total_time_wt;
    }
    if (call_tree->visits == 1)
    {
        call_tree->measurement->wt_total_time += total_time_wt;
    }

    call_tree->method->measurement->pr_self_time += self_time_pr;
    call_tree->method->measurement->pr_wait_time += frame->wait_time.pr;
    call_tree->measurement->pr_self_time += self_time_pr;
    call_tree->measurement->pr_wait_time += frame->wait_time.pr;
    if (call_tree->method->visits == 1)
    {
        call_tree->method->measurement->pr_total_time += total_time_pr;
    }
    if (call_tree->visits == 1)
    {
        call_tree->measurement->pr_total_time += total_time_pr;
    }

    call_tree->method->measurement->al_self_time += self_time_al;
    call_tree->method->measurement->al_wait_time += frame->wait_time.al;
    call_tree->measurement->al_self_time += self_time_al;
    call_tree->measurement->al_wait_time += frame->wait_time.al;
    if (call_tree->method->visits == 1)
    {
        call_tree->method->measurement->al_total_time += total_time_al;
    }
    if (call_tree->visits == 1)
    {
        call_tree->measurement->al_total_time += total_time_al;
    }

    call_tree->method->measurement->mm_self_time += self_time_mm;
    call_tree->method->measurement->mm_wait_time += frame->wait_time.mm;
    call_tree->measurement->mm_self_time += self_time_mm;
    call_tree->measurement->mm_wait_time += frame->wait_time.mm;

    call_tree->method->visits--;
    call_tree->visits--;

    prof_frame_t *parent_frame = prof_stack_last(stack);
    if (parent_frame)
    {
        parent_frame->child_time.wt += total_time_wt;
        parent_frame->child_time.pr += total_time_pr;
        parent_frame->child_time.al += total_time_al;
        parent_frame->child_time.mm += total_time_mm;

        parent_frame->dead_time.wt += frame->dead_time.wt;
        parent_frame->dead_time.pr += frame->dead_time.pr;
        parent_frame->dead_time.al += frame->dead_time.al;
        parent_frame->dead_time.mm += frame->dead_time.mm;
    }

    frame->source_file = Qnil;

    return frame;
}

prof_method_t *prof_find_method(prof_stack_t *stack, VALUE source_file, int source_line)
{
    prof_frame_t *frame = prof_stack_last(stack);

    if (!stack)
    {
        return NULL;
    }

    while (frame >= stack->start)
    {
        if (!frame)
        {
            return NULL;
        }
        if (!source_file)
        {
            return NULL;
        }
        if (!frame->call_tree)
        {
            return NULL;
        }
        if (!frame->call_tree->method)
        {
            return NULL;
        }
        if (!frame->call_tree->method->source_line)
        {
            return NULL;
        }
        if (!frame->call_tree->method->source_file)
        {
            return NULL;
        }

        if (rb_str_equal(source_file, frame->call_tree->method->source_file) && source_line >= frame->call_tree->method->source_line)
        {
            return frame->call_tree->method;
        }
        frame--;
    }
    return NULL;
}
